/*
 * $Id: search.c,v 1.45 2000/02/06 10:37:46 tatuo-y Exp $
 *
 * string search and processing results
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "sufary.h"
#include "util.h"
#include "config.h"

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))


int sa_error_no;
char *sa_error_str[] = {
/* 00 */ "THE ERROR",
/* 01 */ "MEMORY ALLOCATE ERROR",
/* 02 */ "FILE OPEN ERROR",
/* 03 */ "MMAP ERROR",
/* 04 */ "KEYWORD PARSING ERROR"
};


/*
 * struct for binary search
 */
typedef struct sa_bsearch_result_ {
    SA_INDEX center;
    SA_INDEX left;
    SA_INDEX right;
    int left_prefix_len;
    int right_prefix_len;
    int base_txt_skip;
    SA_STAT state;
} SA_BSEARCH_RESULT;


static SA_INDEX m_position_in_string;

/*
 * get position in string after searching
 * this function is not reentrant.
 */
/* SA_INDEX sa_get_position_in_string(void) */
/* { */
/*     return m_position_in_string; */
/* } */


/*
 * set debug mode
 */
void
sa_set_debug_mode(int i)
{
    set_debug_mode(i);
}


/*
 * sistring(Ⱦ̵ʸ) 
 *
 * return value
 *   0 : MATCH 
 *   - : LESS  sistring(pos) < str  ('abc...' < 'ccc')
 *   + : ABOVE sistring(pos) > str  ('abc...' > 'aaa')
 */
static int
cmp_sistr(
    const unsigned char *txt,	/* Ӥʸ(оݥƥ) */
    const int *str,		/* Ӥʸ() */
    int *diffpos,		/* Ϥưۤʤä(ʸ) */
    int len			/* Ӥʸ */
    )
{
    int i;

    assert(txt != NULL && str != NULL);
    assert(0 <= len);

    for (i = 0; i < len; i++) {
	int t = (int)(*txt);
	if (t != *str) {
	    *diffpos = i;
	    sa_dprintf("intcmp: (%d) - (%d) = %d\n", t, *str, (t - *str));
	    return (t - *str);
	}
	txt++;
	str++;
    }
    *diffpos = len;
    return 0;
}


/*
 * binary search called from sa_find
 */
static SA_BSEARCH_RESULT
sa_bsearch(
    const SUFARY *ary,
    SA_BSEARCH_RESULT sbr,
    const int *key,
    int keylen
    )
{
    assert(0 <= sbr.left && sbr.right <= sa_get_array_size(ary));
    assert(sbr.left < sbr.right);

    sbr.state = FAIL;

    sa_dprintf("bsearch: start! [ %ld, %ld )\n", sbr.left, sbr.right);

    while (sbr.left < sbr.right) {  /* avoid negative number (usually "<=")*/
	int txt_skip = MIN(sbr.left_prefix_len, sbr.right_prefix_len);
	int key_skip = txt_skip - sbr.base_txt_skip;
	int diffpos;		/* ʸӤǰۤʤΤä */
 	SA_INDEX cur =  /* avoid overflow */
	    sbr.left / 2 + sbr.right / 2 + (1 & sbr.left & sbr.right);
	int result_of_cmp = cmp_sistr(
	    (unsigned char*)sa_aryidx2txtptr(ary, cur) + txt_skip,
	    key + key_skip,
	    &diffpos, keylen - key_skip);

	sa_dprintf("bsearch: txt_skip = %d, key_skip = %d, base_txt_skip = %d\n",
		(int)txt_skip, (int)key_skip, (int)sbr.base_txt_skip);
	sa_dprintf("bsearch: %ld = ( %ld + %ld ) / 2  cmp = %d\n",
		cur, sbr.left, sbr.right, result_of_cmp);

	if (result_of_cmp < 0) { /* text < key */
	    sbr.left = cur + 1;
	    sbr.left_prefix_len = txt_skip + diffpos;
	} else if (result_of_cmp > 0) {	/* text > key */
	    sbr.right = cur;
	    sbr.right_prefix_len = txt_skip + diffpos;
	} else {		/* text == key */
	    sbr.state = SUCCESS;
	    sbr.center = cur;
	    break;
	}
    }

    if (sbr.state == FAIL)
	sbr.center = sbr.left / 2 + sbr.right / 2 + (1 & sbr.left & sbr.right);

    sa_dprintf("bsearch: * %ld  [ %ld, %ld )\n",
	    sbr.center, sbr.left, sbr.right);
    
    return sbr;
}


/*
 * suffix array search
 *
 * example: searching for "fb" (keyword)
 * results: from l to r (represented by '#')
 *
 *                 abxxiabcdefgxyzpoop
 *   left          hipdabbbbbbbbbbxaba      right
 *    |            bbbcfffffffffffhllm        |
 *    +-----------------##########------------+
 *                      |        |
 *                      l        r
 */
SUF_RESULT
sa_find(
    const SUFARY *ary,
    SA_INDEX left,		/* result-range: left-inside */
    SA_INDEX right,		/* result-range: right-inside */
    const char *key,		/*  */
    int keylen,			/* ʸĹ */
    int base_txt_skip		/* оʸΥå׿ */
    )
{
    static int keyary[KEYWORD_MAX_LENGTH];
    SA_BSEARCH_RESULT sbr, sbr_r, sbr_l;
    SUF_RESULT sr;
    int i;

    assert(0 <= left && right < ary->arraysize);
    assert(base_txt_skip >= 0);

    sa_dprintf("find: key = %.*s\n", keylen, key);
    for (i = 0; i < keylen; i++) {
	keyary[i] = (unsigned char)key[i];
	sa_dprintf("find: key[%d] = '%c', keyary[%d] = %d\n",
		i, key[i], i, keyary[i]);
    }

    sbr.left = left;
    sbr.right = right + 1;
    sbr.left_prefix_len = base_txt_skip;
    sbr.right_prefix_len = base_txt_skip;
    sbr.base_txt_skip = base_txt_skip;

    /* step 1: find a position mathces the key "fb" (represented by '*').
     *
     *                 abxxiabcdefgxyzpoop
     *   left          hipdabbbbbbbbbbxaba      right
     *    |            bbbcfffffffffffhllm        |
     *    +---------------------*-----------------++
     *    |                     |                  |
     *   lb                   center               rb
     */
    sbr = sa_bsearch(ary, sbr, keyary, keylen);
    if (sbr.state == FAIL) {
	sr.stat = FAIL;
	return sr;
    }

    /* step 2: find a right boundary (represented by '*').
     *         '%' means a seach area.
     *
     *                 abxxiabcdefgxyzpoop
     *   left          hipdabbbbbbbbbbxaba      right
     *    |            bbbcfffffffffffhllm        |
     *    +-------------+-------#-----*-+---------+
     *                  |       %%%%%%%%|
     *                  |       |       |
     *                 lb      lor      rb
     */
    keyary[keylen] = 256;
    sbr_r = sbr;
    sbr_r.left = sbr.center;
    sbr_r.left_prefix_len = keylen + base_txt_skip;
    sbr_r = sa_bsearch(ary, sbr_r, keyary, keylen + 1);

    /* step 3: find a left boundary (represented by '*').
     *
     *                 abxxiabcdefgxyzpoop
     *   left          hipdabbbbbbbbbbxaba      right
     *    |            bbbcfffffffffffhllm        |
     *    +-------------+---*---######+-----------+
     *                  %%%%%%%%%     |
     *                  |       |     |
     *                 lb      rol    rb
     */
    keyary[keylen] = -1;
    sbr_l = sbr;
    sbr_l.right = sbr.center + 1;
    sbr_l.right_prefix_len = keylen + base_txt_skip;
    sbr_l = sa_bsearch(ary, sbr_l, keyary, keylen + 1);

    sr.suf = (SUFARY *)ary;
    sr.left = sbr_l.center;
    sr.right = sbr_r.center - 1;
    sr.stat = SUCCESS;
    sa_dprintf("find: found [ %ld, %ld ]\n", sr.left, sr.right);
    return sr;
}


/*
 *   ƥȥեδؿ
 *   - arrayեidxܤǤ֤
 *   - textեidxܤ򤵤ʸݥ󥿤֤
 *   - arrayեidxܤǤʸݥ󥿤֤
 */
SA_INDEX
sa_aryidx2txtidx(const SUFARY *ary, SA_INDEX idx)
{
#ifdef WORDS_BIGENDIAN
    return sa_get_array_ptr(ary)[idx];
#else
    SA_INDEX rv = sa_get_array_ptr(ary)[idx];
    sa_reverse_byte_order(&rv, sizeof(SA_INDEX));
    return rv;
#endif
}

char *
sa_txtidx2txtptr(const SUFARY *ary, SA_INDEX idx)
{
    return sa_get_text_ptr(ary) + idx;
}

char *
sa_aryidx2txtptr(const SUFARY *ary, SA_INDEX idx)
{
#ifdef WORDS_BIGENDIAN
    return sa_get_text_ptr(ary) + sa_get_array_ptr(ary)[idx];
#else
    SA_INDEX rv = sa_get_array_ptr(ary)[idx];
    sa_reverse_byte_order(&rv, sizeof(SA_INDEX));
    return sa_get_text_ptr(ary) + rv;
#endif
}


/*
 *   ΰ踡
 *
 * description
 *   2ʬõˤꡢŪǥå̤κõ
 *
 * () S ϥɥȤγϡEϽλ֤ɽ
 *      target=24 ʤ 2056 Ūɥ
 *         start = 20, size=36  
 *      target=67 ϥɥȳ
 *         start = -1, size=-1  
 *
 * DocID |  0 | 12 | 20 | 56 | 89 | 97 |
 *          S    E    S    E    S    E 
 *
 */
SA_INDEX
sa_dididx2txtidx(const DID *did, SA_INDEX idx)
{
#ifdef WORDS_BIGENDIAN
    return sa_get_did_ptr(did)[idx];
#else
    SA_INDEX rv = sa_get_did_ptr(did)[idx];
    sa_reverse_byte_order(&rv, sizeof(SA_INDEX));
    return rv;
#endif
}

DID_RESULT
sa_didsearch(const DID *did, SA_INDEX target)
{
    SA_INDEX size = sa_get_did_size(did);
    SA_INDEX cur;
    SA_INDEX li = 0;
    SA_INDEX ri = size;
    SA_INDEX region_from = 0;
    SA_INDEX region_to = 0;
    DID_RESULT dr;

    assert(0 <= target);
    sa_dprintf("did-search: target = %ld\n", target);
    while (li < ri) {
	cur = li / 2 + ri / 2 + (1 & li & ri);

	sa_dprintf("did-search: l=[%ld] c=[%ld] r=[%ld]\n", li, cur, ri);
	region_from = sa_dididx2txtidx(did, cur * 2);
	if (region_from <= target) {
	    li = cur + 1;
	} else if (target < region_from) {
	    ri = cur;
	} else
	    break;		/* == */
    }
    
    region_to = sa_dididx2txtidx(did, cur * 2 + 1);

    if (region_from <= target && target < region_to) {
	dr.stat = SUCCESS;
	dr.no = cur;
	dr.start = region_from;
	dr.size = region_to - region_from;
    } else {
	dr.stat = FAIL;
	dr.no = cur + 1;
	if (target < region_from)
	    dr.no = 0;
	dr.start = 0;
	dr.size = 0;
    }

    sa_dprintf("did-search: *** %s [no=%ld start=%ld size=%ld]\n",
	    dr.stat == FAIL ? "NOT-FOUND" : "FOUND",
	    dr.no, dr.start, dr.size);

    return dr;
}

/*
 * access functions for DID
 */
SA_INDEX
sa_get_start_position(const DID *did, SA_INDEX id)
{
    return sa_dididx2txtidx(did, id * 2);
}

SA_INDEX
sa_get_end_position(const DID *did, SA_INDEX id)
{
    return sa_dididx2txtidx(did, id * 2 + 1);
}


/*
 * sequential search for the pattern
 */
SA_INDEX
sa_scan_forward(const SUFARY *ary, SA_INDEX pos, const char *pat)
{
    char *txt = sa_get_text_ptr(ary);
    SA_INDEX txt_size = sa_get_text_size(ary);
    SA_INDEX idx = pos;
    int pat_len = strlen(pat);
    while (idx <= txt_size - pat_len && strncmp(txt + idx, pat, pat_len) != 0)
	idx++;
    idx += pat_len;
    return idx;		/* index = after the pattern */
}

SA_INDEX
sa_scan_backward(const SUFARY *ary, SA_INDEX pos, const char *pat)
{
    char *txt = sa_get_text_ptr(ary);
    SA_INDEX idx = pos;
    int pat_len = strlen(pat);
    while (0 < idx && strncmp(txt + idx, pat, pat_len) != 0)
	idx--;
    return idx;		/* index = start position of the pattern */
}


/*
 * sequential search for start and end tags and get region
 */
SA_INDEXES
sa_get_region_by_scan(const SUFARY *ary, SA_INDEX pos,
		      const char *start_tag, const char *end_tag,
		      int context_b, int context_a)
{
    SA_INDEX from = pos;
    SA_INDEX to = pos;
    SA_INDEXES rv;
    int i;
    assert(0 <= pos && pos < sa_get_text_size(ary));
    assert(0 <= context_a && 0 <= context_b);
    for (i = 0; i <= context_b; i++)
	from = sa_scan_backward(ary, from - (i != 0), start_tag);
    for (i = 0; i <= context_a; i++)
	to = sa_scan_forward(ary, to, end_tag);
    assert(from < to);
    if (from < 0)
	from = 0;
    rv.from = from;
    rv.to = to;
    rv.here = pos;
    return rv;
}


/*
 *   aryβԤǶޤ줿ʸ(Ĥޤꡢ)μ
 */
char *
sa_get_line(const SUFARY *ary, SA_INDEX pos)
{
    return sa_get_lines(ary, pos, 0, 0);
}


/*
 *   aryβԤǶޤ줿ʸ()
 *   ˤιԤƱ˼ä
 */
char *
sa_get_lines(
    const SUFARY *ary,
    SA_INDEX pos,     /* ƥǤΰ */
    int cnt_b,        /* Ʊ˼ä褿ιԤιԿ */
    int cnt_a         /* Ʊ˼ä褿ιԤιԿ */
    )
{
    char *txt = sa_get_text_ptr(ary);
    SA_INDEX txt_size = sa_get_text_size(ary);

    SA_INDEX from = pos;
    SA_INDEX to = pos;

    assert(0 <= pos && pos < txt_size);
    assert(0 <= cnt_a && 0 <= cnt_b);

    while (0 < from && 0 <= cnt_b) {
	from--;
	if (txt[from] == '\n')
	    cnt_b--;
    }
    if (txt[from] == '\n')
	from++;

    while (to < txt_size - 1 && 0 <= cnt_a) {
	to++;
	if (txt[to] == '\n')
	    cnt_a--;
    }
    if (txt[to] == '\n')
	to--;

    m_position_in_string = pos - from;

    {
	SA_INDEX i;
	SA_INDEX size = to - from + 1;
	char *str = malloc(size + 1);

	if (str == NULL) 
	    return NULL; /* malloc fail...  easy error handling */

	/*  4 5 6 7 8 9
	 * \n a b c d \n  -->  str = "abcd\0"
	 *    f     t
	 * strncpy()  \0 ȥԡߤ㤦ΤǻȤʤ */
	for (i = 0; i < size; i++) {
	    str[i] = txt[from + i];
	    /* printf  \0 ʸɽ뤿ζκ */
	    if (str[i] == '\0')
		str[i] = '\1';
	}
	str[size] = '\0';

	return str;
    }
}


/*
 *   ڤʸǶޤ줿ƥȤ
 */
char *
sa_get_block(
    const SUFARY *ary,
    SA_INDEX pos,		/* ƥǤΰ */
    const char *tag1,		/* ڤʸ(start) */
    const char *tag2		/* ڤʸ(end) */
    )
{
    int tag1_len;
    int tag2_len;
    char *str;
    char *txt;
    SA_INDEX txt_size;
    SA_INDEX from = pos;
    SA_INDEX to = pos;

    tag1_len = strlen(tag1);
    tag2_len = strlen(tag2);
    txt_size = sa_get_text_size(ary);
    txt = sa_get_text_ptr(ary);
    assert(0 <= tag1_len && 0 <= tag2_len);
    assert(0 <= pos && pos < txt_size);

    while (0 < from && strncmp(txt + from, tag1, tag1_len) != 0)
	from--;

    while (to <= txt_size - tag2_len && strncmp(txt + to, tag2, tag2_len) != 0)
	to++;

    to += tag2_len;

    m_position_in_string = pos - from;

    assert(from < to);
    str = malloc(to - from + 1);

    if (str == NULL) 
	return NULL; /* malloc fail...  easy error handling */

    strncpy(str, txt + from, to - from);
    str[to - from] = '\0';

    return str;
}


/*
 *   from ʸܤĹ size ʸäƤ
 */
char *
sa_get_string(const SUFARY *ary, SA_INDEX from, SA_INDEX size)
{
    char *str;

    assert(size > 0);
    assert(0 <= from && from < sa_get_text_size(ary));

    str = malloc(size + 1);
    strncpy(str, sa_get_text_ptr(ary) + from, size);
    str[size] = '\0';

    return str;
}

/*
 * array ե뤬ȤƤ뤫ɤĴ٤
 */
int
sa_is_sorted(const SUFARY *ary)
{
  SA_INDEX i;
  for(i = 1; i < sa_get_array_size(ary); i++) 
    if (strcmp(sa_aryidx2txtptr(ary, i - 1), sa_aryidx2txtptr(ary, i)) > 0) {
      sa_dprintf("sort check: not sorted! at %ld\n", i);
      return 0;
    }
  
  sa_dprintf("sort check: sorted!\n");

  return 1;
}
